import quilt3
from karr_lab_aws_manager.config import config
from karr_lab_aws_manager.s3 import util as s3_util
import json
from configparser import ConfigParser
from pathlib import Path, PurePath
import tempfile
from boto3.s3.transfer import TransferConfig


class QuiltUtil:

    def __init__(self, base_path=None, profile_name=None, default_remote_registry=None,
                aws_path=None, cache_dir=None):
        ''' Handle Quilt authentication file creation without having to use quilt3.login()
            Args:
                aws_path (:obj: `str`): directory in which aws credentials file resides
                base_path (:obj: `str`): directory to store quilt3 credentials generated from aws credentials
                profile_name (:obj: `str`): AWS credentials profile name for quilt
                default_remote_registry (:obj: `str`): default remote registry to store quilt package
                cache_dir (:obj: `str`): default directory to store data
        '''
        self.cache_dir = cache_dir
        self.profile_name = profile_name
        base_path_obj = Path(base_path)
        aws_path_obj = Path(aws_path)
        quilt3.session.AUTH_PATH = base_path_obj / 'auth.json'
        quilt3.session.CREDENTIALS_PATH = base_path_obj / 'credentials.json'
        quilt3.session.AUTH_PATH.touch()
        self.auth_path = quilt3.session.AUTH_PATH
        self.quilt_credentials_path = quilt3.session.CREDENTIALS_PATH

        self.aws_credentials_path = aws_path_obj / 'aws_credentials'
        config = ConfigParser()
        config.read(self.aws_credentials_path.expanduser())
        dic = {'access_key': config[profile_name]['aws_access_key_id'],
               'secret_key': config[profile_name]['aws_secret_access_key'],
               'token': None,
               'expiry_time': config[profile_name]['expiry_time']}
        with open(str(self.quilt_credentials_path), 'w') as f:
            json.dump(dic, f)
        quilt3.config(default_remote_registry=default_remote_registry)
        self.package = quilt3.Package()

    def bucket_obj(self, bucket_uri):
        ''' Create quilt3 bucket object
            Args:
                bucket_uri (:obj: `str`): quilt s3 bucket address
            Return:
                (:obj: `quilt3.Bucket`): quilt3 bucket object
        '''
        return quilt3.Bucket(bucket_uri)

    def add_to_package(self, destination=None, source=None, meta=None):
        ''' Specifically used for uploading datanator package to
            quilt3
            Args:
                source (:obj: `list` of :obj: `str`): sources to be added to package,
                                                      directories must end with '/'
                destination (:obj: `list` of :obj: `str` ): package(s) to be manipulated,
                                                            directories must end with '/'
                meta (:obj: `list` of :obj: `dict`): package meta
            Return:

        '''
        length = len(destination)
        if not (all(len(lst)) == length for lst in [source, meta]):
            return 'All three entries must be lists of the same length.'
        suffix = '/'
        for i, d in enumerate(destination):
            s = source[i]
            m = meta[i]
            if s.endswith(suffix) != d.endswith(suffix): # when s and d do not share the same suffix
                return '{} and {} must have the same suffix. Operation stopped at {}th element.'.format(d, s, i)

            if s.endswith(suffix):
                self.package.set_dir(d, s, meta=m)
            else:
                self.package.set(d, s, meta=m)

    def push_to_remote(self, package, package_name, destination=None, message=None) -> str:
        ''' Push local package to remote registry
            Args:
                package (:obj: `quilt3.Package()`): quilt pacakge
                package_name (:obj: `str`): name of package in "username/packagename" format
                destination (:obj: `str`): file landing destination in remote registry
                message (:obj: `str`): commit message
        '''
        try:
            package.push(package_name, dest=destination, message=message)
        except quilt3.util.QuiltException as e:
            return str(e)

    def build_from_external_bucket(self, package_dest, bucket_name, key, file_dir,
                                  bucket_credential=None, profile_name=None, meta=None,
                                  bucket_config=None, max_concurrency=10):
        ''' Build package with source from external (non-quilt)
            s3 buckets
            Args:
                package_dest (:obj: `str`): package(s) to be manipulated
                bucket_name (:obj: `str`): s3 bucket name
                key (:obj: `str`): the name of the key to download from
                file_dir (:obj: `str`): the path to the file to download to
                bucket_credential (:obj: `str`): directory in which credential for s3 bucket is stored
                profile_name (:obj: `str`): profile to be used for authentication
                meta (:obj: `dict`): meta information for package file
                bucket_credential (:obj: `str`): directory in which config for s3 bucket is stored
                max_concurrency (:obj: `int`): threads used for downloading
        '''
        settings = TransferConfig(max_concurrency=max_concurrency)
        if bucket_credential is None:
            bucket_credential = str(self.aws_credentials_path.expanduser())
        else:
            bucket_credential = str(Path(bucket_credential).expanduser())

        if bucket_config is None:
            bucket_config = str(self.aws_credentials_path.with_name('aws_config').expanduser())
        else:
            bucket_config = str(Path(bucket_config).expanduser())

        if profile_name is None:
            profile_name = self.profile_name

        s3 = s3_util.S3Util(profile_name=profile_name, credential_path=bucket_credential,
                            config_path=bucket_config)
                            
        file_name_path = Path(file_dir, key).expanduser()
        file_name = str(file_name_path)
        if package_dest.endswith('/'):
            file_name_path.mkdir(parents=True, exist_ok=True)
            s3.download_dir(key, bucket_name, local=file_dir)
        else:
            Path(file_name_path.parent).mkdir(parents=True, exist_ok=True)
            file_name_path.touch(exist_ok=True)        
            s3.client.download_file(bucket_name, key, file_name, Config=settings)
        
        if package_dest.endswith('/'):
            return self.package.set_dir(package_dest, file_name, meta=meta)
        else:
            return self.package.set(package_dest, file_name, meta=meta)



# def main():
#     manager = QuiltUtil(aws_path='.wc/third_party',
#                         base_path=tempfile.mkdtemp(), profile_name='quilt-s3',
#                         default_remote_registry='s3://karrlab')
#     p = manager.package
#     p = p.set_dir('/', 's3://karrlab/datanator/')
#     p.set_meta({"package-type": 'mongodb data dump'})
#     p.push('karrlab/datanator')

# if __name__ == '__main__':
#     main()


from elasticsearch import Elasticsearch, RequestsHttpConnection
from karr_lab_aws_manager.config import config
import requests
import json
import math
from requests_aws4auth import AWS4Auth


class EsUtil(config.establishES):

    def __init__(self, profile_name=None, credential_path=None,
                config_path=None, elastic_path=None,
                cache_dir=None, service_name='es', max_entries=float('inf'), verbose=False):
        ''' 
            Args:
                profile_name (:obj: `str`): AWS profile to use for authentication
                credential_path (:obj: `str`): directory for aws credentials file
                config_path (:obj: `str`): directory for aws config file
                elastic_path (:obj: `str`): directory for file containing aws elasticsearch service variables
                cache_dir (:obj: `str`): temp directory to store json for bulk upload
                service_name (:obj: `str`): aws service to be used
        '''
        super().__init__(config_path=config_path, profile_name=profile_name, credential_path=credential_path,
                        elastic_path=elastic_path, service_name=service_name)
        self.verbose = verbose
        self.max_entries = max_entries
        self.cache_dir = cache_dir

    def _build_es(self, suffix=None):
        ''' build es query object
            Args:
                suffix (:obj: `str`): string trailing es endpoint
            Returns:
                es (:obj: `Elasticsearch`): Elasticsearch object
        '''
        if suffix is None:
            uri = self.es_endpoint.split('https://')[1]
        else:
            uri = self.es_endpoint.split('https://')[1] + suffix 
        
        es = Elasticsearch(
            hosts = [{'host': uri, 'port': 443}],
            http_auth = self.awsauth,
            use_ssl = True,
            verify_certs = True,
            connection_class = RequestsHttpConnection
        )
        return es
    
    def make_action_and_metadata(self, index, _id):
        ''' Make action_and_metadata obj for bulk loading
            e.g. { "index": { "_index" : "index", "_id" : "id" } }
            Args:
                index (:obj: `str`): name of index on ES
                _id (:obj: `str`):  unique id for document
            Returns:
                action_and_metadata (:obj: `dict`): metadata that conforms to ES bulk load requirement
        '''
        action_and_metadata = {'index': { "_index" : index, "_id" : _id }}
        return action_and_metadata

    def delete_index(self, index, _id=None):
        ''' Delete elasticsearch index
            Args:
                index (:obj: `str`): name of index in es
                _id (:obj: `int`): id of the doc in index (optional)
        '''
        if _id is None:
            url = self.es_endpoint + '/' + index
        else:
            url = self.es_endpoint + '/' + index + '/_doc/' + _id
        r = requests.delete(url, auth=self.awsauth)
        return r.status_code

    def data_to_es_bulk(self, count, cursor, index, bulk_size=100, _id='uniprot_id',
                        headers={ "Content-Type": "application/json" }):
        ''' Load data into elasticsearch service
            Args:
                count (:obj: `int`): cursor size
                cursor (:obj: `pymongo.Cursor` or :obj: `iter`): documents to be PUT to es
                bulk_size (:obj: `int`): number of documents in one PUT
                headers (:obj: `dict`): http header
                _id (:obj: `str`): unique id for identification
            Return:
                status_code (:obj: `set`): set of status codes
        '''
        url = self.es_endpoint + '/_bulk'
        status_code = {201}
        bulk_file = ''
        tot_rounds = math.ceil(count/bulk_size)
        def gen_bulk_file(_id, bulk_file):
            action_and_metadata = self.make_action_and_metadata(index, _id)
            bulk_file += json.dumps(action_and_metadata) + '\n'
            bulk_file += json.dumps(doc) + '\n'  
            return bulk_file          

        for i, doc in enumerate(cursor):
            if i == self.max_entries:
                break
            if self.verbose and i % bulk_size == 0:
                print("Processing bulk {} out of {} ...".format(math.floor(i/bulk_size)+1, tot_rounds))
               
            if i == count - 1:  # last entry
                bulk_file = gen_bulk_file(doc[_id], bulk_file)
                # print('commit last entry')
                # print(bulk_file)
                r = requests.post(url, auth=self.awsauth, data=bulk_file, headers=headers)
                status_code.add(r.status_code)
                return status_code
            elif i % bulk_size != 0 or i == 0: #  bulk_size*(n-1) + 1 --> bulk_size*n - 1
                bulk_file = gen_bulk_file(doc[_id], bulk_file)
                # print('building in between')
                # print(bulk_file)
            else:               # bulk_size * n
                # print('commit endpoint')
                # print(bulk_file)
                r = requests.post(url, auth=self.awsauth, data=bulk_file, headers=headers)
                status_code.add(r.status_code)
                bulk_file = gen_bulk_file(doc[_id], '') # reset bulk_file
                
    def data_to_es_single(self, count, cursor, index, _id='uniprot_id',
                          headers={ "Content-Type": "application/json" }):
        ''' Load data into elasticsearch service
            Args:
                count (:obj: `int`): cursor size
                cursor (:obj: `pymongo.Cursor` or :obj: `iter`): documents to be PUT to es
                es_endpoint (:obj: `str`): elasticsearch endpoint
                headers (:obj: `dict`): http header information
            Return:
                status_code (:obj: `set`): set of status codes
        '''
        url_root = self.es_endpoint + '/' + index + '/_doc/'
        status_code = {201}
        for i, doc in enumerate(cursor):
            if i == self.max_entries:
                break
            if i % 20 == 0 and self.verbose:
                print("Processing doc {} out of {}...".format(i, min(count, self.max_entries)))
            url = url_root + doc[_id]
            r = requests.post(url, auth=self.awsauth, json=doc, headers=headers)
            status_code.add(r.status_code)
        return status_code

# def main():
#     manager = EsUtil()
#     manager.delete_index('protein')

# if __name__ == "__main__":
#     main()


""" karr_lab_aws_manager command line interface

:Author: Name <email>
:Date: 2019-9-16
:Copyright: 2019, Karr Lab
:License: MIT
"""

import cement
import karr_lab_aws_manager
import karr_lab_aws_manager.core


class BaseController(cement.Controller):
    """ Base controller for command line application """

    class Meta:
        label = 'base'
        description = "karr_lab_aws_manager"
        arguments = [
            (['-v', '--version'], dict(action='version', version=karr_lab_aws_manager.__version__)),
        ]

    @cement.ex(help='command_1 description')
    def cmd1(self):
        """ command_1 description """
        print('command_1 output')

    @cement.ex(help='command_2 description')
    def cmd2(self):
        """ command_2 description """
        print('command_2 output')

    @cement.ex(hide=True)
    def _default(self):
        self._parser.print_help()


class Command3WithArgumentsController(cement.Controller):
    """ Command3 description """

    class Meta:
        label = 'command-3'
        description = 'Command3 description'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['arg_1'], dict(
                type=str, help='Description of arg_1')),
            (['arg_2'], dict(
                type=str, help='Description of arg_2')),
            (['--opt-arg-3'], dict(
                type=str, default='default value of opt-arg-1', help='Description of opt-arg-3')),
            (['--opt-arg-4'], dict(
                type=float, default=float('nan'), help='Description of opt-arg-4')),
        ]

    @cement.ex(hide=True)
    def _default(self):
        args = self.app.pargs
        args.arg_1
        args.arg_2
        args.opt_arg_3
        args.opt_arg_4


class App(cement.App):
    """ Command line application """
    class Meta:
        label = 'karr_lab_aws_manager'
        base_controller = 'base'
        handlers = [
            BaseController,
            Command3WithArgumentsController,
        ]

def main():
    with App() as app:
        app.run()


from datanator_query_python.query import query_protein
from datanator_query_python.config import config as config_mongo
from karr_lab_aws_manager.elasticsearch import util


class MongoToES(util.EsUtil):

    def __init__(self, profile_name=None, credential_path=None,
                config_path=None, elastic_path=None,
                cache_dir=None, service_name='es', index=None, max_entries=float('inf'), verbose=False):
        ''' 
            Args:
                profile_name (:obj: `str`): AWS profile to use for authentication
                credential_path (:obj: `str`): directory for aws credentials file
                config_path (:obj: `str`): directory for aws config file
                elastic_path (:obj: `str`): directory for file containing aws elasticsearch service variables
                cache_dir (:obj: `str`): temp directory to store json for bulk upload
                service_name (:obj: `str`): aws service to be used
        '''
        super().__init__(profile_name=profile_name, credential_path=credential_path,
                config_path=config_path, elastic_path=elastic_path,
                cache_dir=cache_dir, service_name=service_name, max_entries=max_entries, verbose=verbose)
        self.index = index


    def data_from_mongo_protein(self, server, db, username, password, verbose=False,
                                readPreference='nearest', authSource='admin', projection={'_id': 0},
                                query={}):
        ''' Acquire documents from protein collection in datanator
            Args:
                server (:obj: `str`): mongodb ip address
                db (:obj: `str`): database name
                username (:obj: `str`): username for mongodb login
                password (:obj: `str`): password for mongodb login
                verbose (:obj: `bool`): display verbose messages
                readPreference (:obj: `str`): mongodb readpreference
                authSource (:obj: `str`): database login info is authenticating against
                projection (:obj: `str`): mongodb query projection
                query (:obj: `str`): mongodb query filter
            Return:
                docs (:obj: `pymongo.Cursor`): pymongo cursor object that points to all documents in protein collection
                count (:obj: `int`): number of documents returned
        '''
        protein_manager = query_protein.QueryProtein(server=server, database=db,
                 verbose=verbose, username=username, authSource=authSource,
                 password=password, readPreference=readPreference)
        docs = protein_manager.collection.find(filter=query, projection=projection)
        count = protein_manager.collection.count_documents(query)
        return (count, docs)


def main():
    conf = config_mongo.Config()
    username = conf.USERNAME
    password = conf.PASSWORD
    server = conf.SERVER
    authDB = conf.AUTHDB
    db = 'datanator'
    manager = MongoToES(verbose=True)
    
    # data from "protein" collection
    count, docs = manager.data_from_mongo_protein(server, db, username, password, authSource=authDB)
    status_code = manager.data_to_es_bulk(count, docs) 
    
    print(status_code)   

if __name__ == "__main__":
    main()